/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.filter.ooxml.xlsx;

import java.io.InputStream;
import java.util.Map;
import org.apache.commons.io.IOUtils;
import org.json.JSONObject;
import org.xlsx4j.jaxb.Context;
import com.openexchange.office.filter.api.DocumentProperties;
import com.openexchange.office.filter.api.IImporter;
import com.openexchange.office.filter.api.IPartImporter;
import com.openexchange.office.filter.ooxml.OfficeOpenXMLOperationDocument;
import com.openexchange.session.Session;

/**
 * {@link Importer}
 *
 * @author <a href="mailto:sven.jacobi@open-xchange.com">Sven Jacobi</a>
 */
public class Importer implements IImporter, IPartImporter {

    /**
     * Initializes a new {@link Importer}.
     */
    public Importer() {
        Context.getsmlObjectFactory();
    }

    @Override
    public JSONObject createOperations(Session session, InputStream inputDocumentStm, DocumentProperties documentProperties) {
        boolean debugGetDefaultDocument = false;
        if(debugGetDefaultDocument) {   // just a simple helper method to store our in memory created
            try(XlsxOperationDocument emptyDoc = new XlsxOperationDocument(null, null, documentProperties)) {
                documentProperties.put(DocumentProperties.PROP_INITIAL_SHEETNAME, "empty_sheet");
                emptyDoc.createPackage();
                emptyDoc.writeFile("/tmp/default.oxt");    // template documents...
            }
            catch(Throwable e) {
                OfficeOpenXMLOperationDocument.rethrowFilterException(e, null);
            }
        }

        JSONObject operations = null;
        try(XlsxOperationDocument operationDocument = new XlsxOperationDocument(session, null, documentProperties)) {
            operationDocument.loadDocument(inputDocumentStm, true);
            operations = operationDocument.getOperations();
        }
        catch(Throwable e) {
            OfficeOpenXMLOperationDocument.rethrowFilterException(e, null);
        }
        return operations;
    }

    @Override
    public InputStream getDefaultDocument(InputStream templateDocument, DocumentProperties documentProperties) {
        InputStream defaultDocument = null;
        try(XlsxOperationDocument operationDocument = new XlsxOperationDocument(null, null, documentProperties)) {
            if(templateDocument!=null) {
                operationDocument.loadDocument(templateDocument, false);
            }
            else {
                operationDocument.createPackage();
            }
            operationDocument.applyDefaultDocumentProperties();
            defaultDocument = operationDocument.save();
        }
        catch(Throwable e) {
            OfficeOpenXMLOperationDocument.rethrowFilterException(e, null);
        }
        finally {
            IOUtils.closeQuietly(templateDocument);
        }
        return defaultDocument;
    }

    @Override
    public void initPartitioning(Session session, InputStream inputDocument, DocumentProperties documentProperties) {
        try(XlsxOperationDocument operationDocument = new XlsxOperationDocument(session, null, documentProperties)) {
            operationDocument.loadDocument(inputDocument, true);
            documentProperties.put(DocumentProperties.PROP_DOCUMENT, operationDocument);
        }
        catch(Throwable e) {
            OfficeOpenXMLOperationDocument.rethrowFilterException(e, null);
        }
    }

    @Override
    public Map<String, Object> getMetaData(DocumentProperties documentProperties) {
        Map<String, Object> metaData = null;
        final XlsxOperationDocument operationDocument = (XlsxOperationDocument)documentProperties.get(DocumentProperties.PROP_DOCUMENT);
        try {
            operationDocument.registerMemoryListener();
            metaData = operationDocument.getMetaData();
        }
        catch(Throwable e) {
            OfficeOpenXMLOperationDocument.rethrowFilterException(e, operationDocument.getPackage());
        }
        finally {
            operationDocument.close();
        }
        return metaData;
    }

    @Override
    public Map<String, Object> getActivePart(DocumentProperties documentProperties) {
        Map<String, Object> metaData = null;
        final XlsxOperationDocument operationDocument = (XlsxOperationDocument)documentProperties.get(DocumentProperties.PROP_DOCUMENT);
        try {
            operationDocument.registerMemoryListener();
            metaData = operationDocument.getActivePart();
        }
        catch(Throwable e) {
            OfficeOpenXMLOperationDocument.rethrowFilterException(e, operationDocument.getPackage());
        }
        finally {
            operationDocument.close();
        }
        return metaData;
    }

    @Override
    public Map<String, Object> getNextPart(DocumentProperties documentProperties) {
        Map<String, Object> metaData = null;
        final XlsxOperationDocument operationDocument = (XlsxOperationDocument)documentProperties.get(DocumentProperties.PROP_DOCUMENT);
        try {
            operationDocument.registerMemoryListener();
            metaData = operationDocument.getNextPart();
        }
        catch(Throwable e) {
            OfficeOpenXMLOperationDocument.rethrowFilterException(e, operationDocument.getPackage());
        }
        finally {
            operationDocument.close();
        }
        return metaData;
    }
}
