/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
package com.openexchange.office.session;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.Validate;
import com.openexchange.context.ContextService;
import com.openexchange.exception.OXException;
import com.openexchange.groupware.contexts.Context;
import com.openexchange.office.tools.osgi.ServiceLookupRegistry;
import com.openexchange.office.tools.osgi.ServiceLookupRegistryService;
import com.openexchange.session.Session;
import com.openexchange.sessiond.SessiondService;
import com.openexchange.sessiond.SessiondServiceExtended;
import com.openexchange.tools.session.ServerSession;
import com.openexchange.tools.session.ServerSessionAdapter;
import com.openexchange.user.User;
import com.openexchange.user.UserService;

//=============================================================================
/** helper to handle seesions and it's values more comfortable.
 */
public class SessionService
{
	//-------------------------------------------------------------------------
	/** @return a session object bound to the given session ID.
	 *          Will return null if session ID is invalid, unknown, wrong, unexpected ...
	 *
	 *	@param	sSessionId [IN]
	 *			the session ID where such sessions has to be returned for.
	 * @throws OXException
	 */
	@SuppressWarnings("unchecked")
	public < T extends Session > T getSession4Id (final String sSessionId) throws OXException
	{
		if (StringUtils.isEmpty(sSessionId))
			return null;

		final SessiondService       aSessionD = ServiceLookupRegistryService.getInstance().getService(SessiondService.class);
		      Session               aSession  = null;

		if (null == aSessionD) {
			throw new RuntimeException("Mandatory service 'SessiondService' not available therefore unable to proceed, check your installation");
		}

		if (aSessionD instanceof SessiondServiceExtended)
		{
			final SessiondServiceExtended aSessionDEx = (SessiondServiceExtended) aSessionD;
			aSession = aSessionDEx.getSession(sSessionId);
		}

		if (aSession == null)
		{
			aSession = aSessionD.getSession(sSessionId);
		}

		final ServerSession aSessionEx = ServerSessionAdapter.valueOf(aSession);
		return (T) aSessionEx;
	}

	//-------------------------------------------------------------------------
	/** @return TRUE if given session is valid and not expired; FALSE otherwise.
	 *
	 *	@param	aSession [IN]
	 *			the session to be validated here.
	 * @throws OXException
	 */
	public < T extends Session > boolean isSessionValid (final T aSession) throws OXException
	{
		if (aSession == null)
			return false;

		final Context aContext = getContextInfo(aSession);
        if (
        	(   aContext == null    ) ||
        	( ! aContext.isEnabled())
           )
        	return false;

        final User aUser = getUserInfo(aSession);
        if (
        	(   aUser == null        ) ||
        	( ! aUser.isMailEnabled())
           )
        	return false;

        return true;
	}

	//-------------------------------------------------------------------------
	@SuppressWarnings("unused")
	public < T extends Session > void refreshSession (final T aSession) throws OXException
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");
		final String  sSessionId        = aSession.getSessionID();
		final Session aRefreshedSession = getSession4Id (sSessionId);
	}

	//-------------------------------------------------------------------------
	/** extract unique session ID from given session.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those session ID has to be retrieved from.
	 *
	 *  @return the unique session ID.
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or session ID could not be retrieved successfully.
	 */
	public < T extends Session > String getSessionId (final T aSession)
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");
		final String sId = aSession.getSessionID();
		Validate.notEmpty(sId, "Invalid session ID.");
		return sId;
	}

	//-------------------------------------------------------------------------
	/** extract unique user ID from given session.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those user ID has to be retrieved from.
	 *
	 *  @return the unique user ID.
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or user ID could not be retrieved successfully.
	 */
	public < T extends Session > int getUserId (final T aSession)
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");
		final int nUserId = aSession.getUserId();
		return nUserId;
	}

	//-------------------------------------------------------------------------
	/** extract unique (user) context ID from given session.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those context ID has to be retrieved from.
	 *
	 *  @return the unique context ID.
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or context ID could not be retrieved successfully.
	 */
	public < T extends Session > int getContextId (final T aSession)
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");
		final int nContextId = aSession.getContextId();
		return nContextId;
	}

	//-------------------------------------------------------------------------
	/** extract unique user ID from given session.
	 *  and return it's string representation.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those user ID has to be retrieved from.
	 *
	 *  @return the unique user ID as string.
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or user ID could not be retrieved successfully.
	 */
	public < T extends Session > String getUserIdString (final T aSession)
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");
		final int    nUserId = getUserId(aSession);
		final String sUserId = Integer.toString(nUserId);
		return sUserId;
	}

	//-------------------------------------------------------------------------
	/** extract unique (user) context ID from given session
	 *  and return it's string representation.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those context ID has to be retrieved from.
	 *
	 *  @return the unique context ID as string.
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or context ID could not be retrieved successfully.
	 */
	public < T extends Session > String getContextIdString (final T aSession)
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");
		final int    nContextId = getContextId(aSession);
		final String sContextId = Integer.toString(nContextId);
		return sContextId;
	}

	//-------------------------------------------------------------------------
	/** analyze the given session and return an User object which
	 *  describe the user more in it's details.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those user info has to be retrieved from.
	 *
	 *  @return the user info object.
	 * @throws OXException
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or user info could not be retrieved successfully.
	 */
	public < T extends Session > User getUserInfo (final T aSession) throws OXException
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");

		final ServiceLookupRegistryService aLookup         = ServiceLookupRegistryService.getInstance();
    	final UserService           aUserService    = aLookup.getService(UserService   .class);
    	final ContextService        aContextService = aLookup.getService(ContextService.class);

		Validate.notNull(aUserService   , "Miss service '"+UserService   .class.getName ()+"'.");
		Validate.notNull(aContextService, "Miss service '"+ContextService.class.getName ()+"'.");

		final int  nUserId    = getUserId    (aSession);
		final int  nContextId = getContextId (aSession);
		final User aUserInfo  = aUserService.getUser(nUserId, nContextId);

		return aUserInfo;
	}

	//-------------------------------------------------------------------------
	/**
     *  Analyze the given session and return an User object which
     *  describe the user more in it's details.
	 *
     * @param  aSession [IN]
     *         the session where those user info has to be retrieved from.
     *
	 * @param nUserId [IN]
	 *         a user id that should be used to retrieve the user data.
	 *
     * @return the user info object.
	 * @throws OXException
     *
     * @throws an exception in case session is not valid ...
     *         or user info could not be retrieved successfully.
	 */
    public < T extends Session > User getUserInfo (final T   aSession,
                                                          final int nUserId ) throws OXException
    {
        Validate.notNull(aSession, "Invalid argument 'session'.");

		final ServiceLookupRegistry aLookup         = ServiceLookupRegistry.get();
        final UserService           aUserService    = aLookup.getService(UserService   .class);
        final ContextService        aContextService = aLookup.getService(ContextService.class);

        Validate.notNull(aUserService   , "Miss service '"+UserService   .class.getName ()+"'.");
        Validate.notNull(aContextService, "Miss service '"+ContextService.class.getName ()+"'.");

        final int  nContextId = getContextId (aSession);
        final User aUserInfo  = aUserService.getUser(nUserId, nContextId);

        return aUserInfo;
    }

    //-------------------------------------------------------------------------
	/** analyze the given session and return an Context object which
	 *  describe the user context more in it's details.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those context info has to be retrieved from.
	 *
	 *  @return the context info object.
	 * @throws OXException
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or user context could not be retrieved successfully.
	 */
	public < T extends Session > Context getContextInfo (final T aSession) throws OXException
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");

    	final ContextService aContextService = ServiceLookupRegistry.get().getService(ContextService.class);

		Validate.notNull(aContextService, "Miss service '"+ContextService.class.getName ()+"'.");

		final int     nContextId = getContextId (aSession  );
		final Context aContext   = aContextService.getContext(nContextId);

		return aContext;
	}

	//-------------------------------------------------------------------------
	/** extract users name from given session to be used as visible information
	 *  e.g. at the UI.
	 *
	 *  @param	aSession [IN]
	 *  		the session where those user name has to be retrieved from.
	 *
	 *  @return the user name.
	 * @throws OXException
	 *
	 *  @throws an exception in case session is not valid ...
	 *  		or user name could not be retrieved successfully.
	 */
	public < T extends Session > String getUserDisplayName (final T aSession) throws OXException
	{
		Validate.notNull(aSession, "Invalid argument 'session'.");

		final User   aUser = getUserInfo(aSession);
		final String sName = aUser.getDisplayName();

		Validate.notEmpty(sName, "Was not able to retrieve users display name from given session.");
		return sName;
	}

	//-------------------------------------------------------------------------
    /** dump given session to string for debug purposes.
    *
    *   @param	aSession [IN]
    *  			the session to be analyzed here
    *
    *   @return the string representation of those session.
    */
    public < T extends Session > String dumpSessionToString (final T aSession)
		// no exception here please .-)
    {
        final StringBuffer sDump = new StringBuffer(256);

        sDump.append("#### =================================================\n");
        sDump.append("#### session dump :                                   \n");
        try {
            sDump.append("#### session      : " + aSession + "\n");
        } catch (final Throwable ignore) {
        } // catch errors line by line ! we want to get info's as much as possible :-)
        try {
            sDump.append("#### login        : " + aSession.getLogin() + "\n");
        } catch (final Throwable ignore) {
        }
        try {
            sDump.append("#### login name   : " + aSession.getLoginName() + "\n");
        } catch (final Throwable ignore) {
        }
        try {
            sDump.append("#### user login   : " + aSession.getUserlogin() + "\n");
        } catch (final Throwable ignore) {
        }
        try {
            sDump.append("#### user id      : " + aSession.getUserId() + "\n");
        } catch (final Throwable ignore) {
        }
        try {
            sDump.append("#### context id   : " + aSession.getContextId() + "\n");
        } catch (final Throwable ignore) {
        }
        sDump.append("#### =================================================\n");

        return sDump.toString();
    }
}
